/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      OnClick
Version:   1.0
ID:        ID_ONCLICK
Status:    Stable
Date:      February 2003
Category:  GUI
Include:   input/onclick.h
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1999-2003.  All rights reserved.
Keywords:  click, button, mouse, point, pointer, buttons
Short:     Manages interface reactions to mouse click events.

###DESCRIPTION###
<p>The OnClick class is used for creating interactive areas between
the user and the display, and is typically facilitated by a pointing device
that supports buttons.  If you need to create clickable icons and buttons on
the display, you need to use this class to actually grab and process the
clicks.</p>

<p>Here is an example of a rendered area that destroys itself when you click
on it:</p>

<pre>
  &lt;render name="drawable" width="40" height="20" colour="#808080"&gt;
    &lt;onclick release&gt;
      &lt;action static call="free" object="[drawable]"/&gt;
    &lt;/onclick&gt;
  &lt/render&gt;
</pre>

<p>The OnClick class is almost always used in conjunction with child objects
that perform a certain action for you.  Typically you will want to use Action
and Script objects to achieve the effect that you require.  Make sure that
you declare them as static, or they will not be initialised to the onclick
object correctly.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/input/joyport.h>
#include <pandora/input/onclick.h>
#include <pandora/modules/screen.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_ONCLICK  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ONCLICK, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1999-2003.  All rights reserved.",
   "June 2003",
   "OnClick"
};

static struct KernelBase *KernelBase;
static struct ScreenBase *ScreenBase;
static OBJECTPTR modScreen = NULL;
static OBJECTPTR OnClickClass = 0;

/*****************************************************************************
** OnClick class definition.
*/

#define LEN_POINTER 28

struct OnClick {
   OBJECT_HEADER           /* Class header */
   LONG     Frame;         /* The frame to change the Render object to when a click occurs */
   LONG     Button;        /* Number of the button to check */
   LONG     ReleaseFrame;  /* The frame to change the Render object to when a click is released */
   LONG     Flags;         /* Special flags */
   OBJECTID DrawableID;    /* Drawable object to be monitored */
   LONG     ClickLimit;    /* Total number of clicks allowed before the object auto-disables itself */
   STRING   ToolTip;       /* Tool tip to show if the pointer sits over the onclick area */
   FLOAT    Delay;         /* Delay for tool tips and auto-activation, measured in seconds */

   PRIVATE_FIELDS
   OBJECTID ToolTipID;
   FLOAT    XCoord;
   FLOAT    YCoord;
   FLOAT    Width;
   FLOAT    Height;
   FLOAT    YOffset;
   FLOAT    XOffset;
   LONG     Dimensions;
   ULONG    MotionTimer;
   UBYTE    Pointer[LEN_POINTER];
   UBYTE    State;
   UBYTE    PointerLocked;
};

static ERROR GET_Height(struct OnClick *, struct Variable *);
static ERROR GET_Pointer(struct OnClick *, STRING *);
static ERROR GET_Width(struct OnClick *, struct Variable *);
static ERROR GET_X(struct OnClick *, struct Variable *);
static ERROR GET_XOffset(struct OnClick *, struct Variable *);
static ERROR GET_Y(struct OnClick *, struct Variable *);
static ERROR GET_YOffset(struct OnClick *, struct Variable *);

static ERROR SET_Height(struct OnClick *, struct Variable *);
static ERROR SET_Pointer(struct OnClick *, STRING);
static ERROR SET_ReleaseFrame(struct OnClick *, LONG *);
static ERROR SET_ToolTip(struct OnClick *, STRING);
static ERROR SET_Width(struct OnClick *, struct Variable *);
static ERROR SET_X(struct OnClick *, struct Variable *);
static ERROR SET_XOffset(struct OnClick *, struct Variable *);
static ERROR SET_Y(struct OnClick *, struct Variable *);
static ERROR SET_YOffset(struct OnClick *, struct Variable *);

static struct FieldDef OnClickFlags[] = {
   { "HELDRELEASE",    OCF_HELDRELEASE    }, { "RELEASE",      OCF_RELEASE },
   { "RELEASEFRAME",   OCF_RELEASEFRAME   }, { "DBLCLICK",     OCF_DOUBLECLICK },
   { "DOUBLECLICK",    OCF_DOUBLECLICK    }, { "DISABLED",     OCF_DISABLED },
   { "TOGGLEFRAME",    OCF_TOGGLEFRAME    }, { "MESSAGE",      OCF_MESSAGE },
   { "IGNORECHILDREN", OCF_IGNORECHILDREN }, { "IGNOREDOUBLE", OCF_IGNOREDOUBLE },
   { "AUTOACTIVATE",   OCF_AUTOACTIVATE   }, { "AUTOLIMIT",    OCF_AUTOLIMIT },
   { "ACTIVATED",      OCF_ACTIVATED },
   { NULL, NULL }
};

static struct FieldArray OnClickFields[] = {
   { "Frame",        0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Button",       0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ReleaseFrame", 0, FDF_LONG|FDF_RW,      0, NULL, SET_ReleaseFrame },
   { "Flags",        0, FDF_LONGFLAGS|FDF_RW, (LONG)&OnClickFlags, NULL, NULL },
   { "Drawable",     0, FDF_OBJECTID|FDF_RI,  ID_RENDER, NULL, NULL },
   { "ClickLimit",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ToolTip",      0, FDF_STRING|FDF_RW,    0, NULL, SET_ToolTip },
   { "Delay",        0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Height",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Pointer",     0, FDF_STRING|FDF_RW,                             0, GET_Pointer, SET_Pointer },
   { "Width",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR ONCLICK_ActionNotify(struct OnClick *, struct acActionNotify *);
static ERROR ONCLICK_Activate(struct OnClick *, APTR);
static ERROR ONCLICK_ClosingTag(struct OnClick *, APTR);
static ERROR ONCLICK_Disable(struct OnClick *, APTR);
static ERROR ONCLICK_Enable(struct OnClick *, APTR);
static ERROR ONCLICK_Free(struct OnClick *, APTR);
static ERROR ONCLICK_NewObject(struct OnClick *, APTR);
static ERROR ONCLICK_Init(struct OnClick *, APTR);
static ERROR ONCLICK_Timer(struct OnClick *, struct acTimer *);

static struct ActionArray OnClickActions[] = {
   { AC_ActionNotify, ONCLICK_ActionNotify },
   { AC_Activate,     ONCLICK_Activate },
   { AC_ClosingTag,   ONCLICK_ClosingTag },
   { AC_Disable,      ONCLICK_Disable },
   { AC_Enable,       ONCLICK_Enable },
   { AC_Free,         ONCLICK_Free },
   { AC_Init,         ONCLICK_Init },
   { AC_NewObject,    ONCLICK_NewObject },
   { AC_Timer,        ONCLICK_Timer },
   { NULL, NULL }
};

enum {
   STATE_ENTERED=1,
   STATE_INSIDE,
   STATE_EXITED
};

static void SubscribeDrawable(struct OnClick *, OBJECTID);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modScreen, NULL,
      FID_Name|TSTRING,   "screen",
      FID_Version|TFLOAT, MODVERSION_SCREEN,
      TAGEND) IS ERR_Okay) {
      if (GetField(modScreen, FID_ModBase, FT_POINTER, &ScreenBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&OnClickClass, NULL,
      FID_BaseClassID|TLONG, ID_ONCLICK,
      FID_Name|TSTRING,      "OnClick",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      OnClickActions,
      FID_Fields|TPTR,       OnClickFields,
      FID_Size|TLONG,        sizeof(struct OnClick),
      FID_Version|TFLOAT,    VER_ONCLICK,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (OnClickClass) { acFree(OnClickClass); OnClickClass = NULL; }
   if (modScreen)    { acFree(modScreen);    modScreen = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** OnClick: ActionNotify
*/

static ERROR ONCLICK_ActionNotify(struct OnClick *Self, struct acActionNotify *NotifyArgs)
{
   if ((NotifyArgs->ActionID IS AC_Free) AND (NotifyArgs->Error IS ERR_Okay)) {
      Action(AC_Free, Self, NULL);
   }
   else if ((NotifyArgs->ActionID IS AC_UserClick) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserClick *Click;
      OBJECTPTR drawable;
      LONG width, height, xcoord, ycoord, frame;

      if (Self->Flags & OCF_DISABLED) return(ERR_Okay);

      Click = (struct acUserClick *)NotifyArgs->Args;

      /*** User Click Handling ***/

      switch(Self->Button) {
         case 1  : if (!(Click->ButtonFlags & JD_LMB)) return(ERR_Okay); break;
         case 2  : if (!(Click->ButtonFlags & JD_RMB)) return(ERR_Okay); break;
         case 3  : if (!(Click->ButtonFlags & JD_MMB)) return(ERR_Okay); break;
         default : return(ERR_Failed);
      }

      /* If this object only supports double clicks, return immediately (double
      ** clicks are only processed when the second-click is released - see
      ** UserClickRelease support).
      */

      if (Self->Flags & OCF_DOUBLECLICK) return(ERR_Okay);

      /*** Ascertain the monitored area ***/

      if (Self->Dimensions) {
         GetFields(Self, FID_Width|TLONG,  &width,
                         FID_Height|TLONG, &height,
                         FID_XCoord|TLONG, &xcoord,
                         FID_YCoord|TLONG, &ycoord,
                         TAGEND);

         if (Click->XCoord < xcoord) return(ERR_Okay);
         if (Click->YCoord < ycoord) return(ERR_Okay);
         if (Click->XCoord >= xcoord + width) return(ERR_Okay);
         if (Click->YCoord >= ycoord + height) return(ERR_Okay);
      }

      /*** Update the rendered frame if required ***/

      if ((Self->Frame) AND (Self->DrawableID)) {
         if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {

            /*** Switch frames if the TOGGLEFRAME flag has been set ***/

            if (Self->Flags & OCF_TOGGLEFRAME) {
               GetField(drawable, FID_Frame, FT_LONG, &frame);

               if (frame IS Self->Frame) {
                  Self->Frame = Self->ReleaseFrame;
                  Self->ReleaseFrame = frame;
               }
               else if (frame IS Self->ReleaseFrame) {
                  if (frame != Self->Frame) {
                     /* The onclick frame should always match the drawable's frame, so reverse them */
                     frame = Self->Frame;
                     Self->Frame = Self->ReleaseFrame;
                     Self->ReleaseFrame = frame;
                     frame = Self->Frame;
                  }

                  frame = Self->ReleaseFrame;
                  Self->ReleaseFrame = Self->Frame;
                  Self->Frame = frame;
               }
            }

            SetField(drawable, FID_Frame, FT_LONG, &Self->Frame);
            ReleaseObject(drawable);

            DelayMsg(AC_Draw, Self->DrawableID, NULL);
         }
         else DPrintF("UserClick:","[OnClick] Failed to access drawable #%d.", Self->DrawableID);
      }

      if (!(Self->Flags & OCF_RELEASE)) acActivate(Self);

      Self->Flags |= OCF_CLICKED;
   }
   else if ((NotifyArgs->ActionID IS AC_UserClickRelease) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct acUserClickRelease *Release;
      OBJECTPTR drawable;
      LONG xcoord, ycoord, width, height;

      if ((Self->Flags & OCF_DISABLED) AND (!(Self->Flags & OCF_CLICKED))) return(ERR_Okay);

      Release = (struct acUserClickRelease *)NotifyArgs->Args;

      /*** Click Release Handling ***/

      switch(Self->Button) {
         case 1  : if (!(Release->ButtonFlags & JD_LMB)) return(ERR_Okay); break;
         case 2  : if (!(Release->ButtonFlags & JD_RMB)) return(ERR_Okay); break;
         case 3  : if (!(Release->ButtonFlags & JD_MMB)) return(ERR_Okay); break;
         default : return(ERR_Failed);
      }

      Self->Flags &= ~OCF_CLICKED;

      /*** Restore the pointer to normal if we have locked it and the mouse is outside of our area ***/

      if ((Self->State != STATE_ENTERED) AND (Self->State != STATE_INSIDE)) {
         if (Self->PointerLocked) {
            RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
            Self->PointerLocked = FALSE;
         }
      }

      /* If this object only supports double clicks, check if a
      ** double click has been sent through the UserClickRelease action.
      */

      if ((Self->Flags & OCF_DOUBLECLICK) AND (!(Release->ButtonFlags & JD_DBLCLICK))) {
         return(ERR_Okay);
      }

      /*** If the IGNOREDOUBLE flag is set, do not process double-clicks ***/

      if (Self->Flags & OCF_IGNOREDOUBLE) {
         if (Release->ButtonFlags & JD_DBLCLICK) {
            return(ERR_Okay);
         }
      }

      /*** Ascertain the monitored area ***/

      if (Self->Dimensions) {
         GetFields(Self, FID_Width|TLONG,  &width,
                         FID_Height|TLONG, &height,
                         FID_XCoord|TLONG, &xcoord,
                         FID_YCoord|TLONG, &ycoord,
                         TAGEND);

         if (Release->XCoord < xcoord) return(ERR_Okay);
         if (Release->YCoord < ycoord) return(ERR_Okay);
         if (Release->XCoord >= xcoord + width) return(ERR_Okay);
         if (Release->YCoord >= ycoord + height) return(ERR_Okay);
      }

      /*** Change the drawable object's graphical frame if necessary ***/

      if ((Self->DrawableID) AND (!(Self->Flags & OCF_TOGGLEFRAME))) {
         if ((Self->Frame) OR (Self->ReleaseFrame)) {
            if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
               SetField(drawable, FID_Frame, FT_LONG, &Self->ReleaseFrame);
               ReleaseObject(drawable);

               DelayMsg(AC_Draw, Self->DrawableID, NULL);
            }
         }
      }

      if ((Self->Flags & (OCF_RELEASE|OCF_HELDRELEASE|OCF_DOUBLECLICK))) {
         /* Figure out whether this will be a held-release (the user held the click
         ** and moved the pointer) or a click-release (the user released the click
         ** at approximately the same position as it was initially made).
         */

         if ((Release->XChange < -2) OR (Release->XChange > 2) OR
             (Release->YChange > 2) OR (Release->YChange < -2)) {
            if (Self->Flags & OCF_HELDRELEASE) Action(AC_Activate, Self, NULL);
         }
         else if (Self->Flags & (OCF_RELEASE|OCF_DOUBLECLICK)) Action(AC_Activate, Self, NULL);
      }

      if (Self->ClickLimit > 0) {
         Self->ClickLimit--;
         if (!Self->ClickLimit) Self->Flags |= OCF_DISABLED;
      }
   }
   else if ((NotifyArgs->ActionID IS AC_UserMovement) AND (NotifyArgs->Error IS ERR_Okay)) {

      if ((Self->Pointer[0]) OR (Self->ToolTip) OR (Self->Flags & OCF_AUTOACTIVATE)) {
         struct acUserMovement *Args;
         LONG width, height, xcoord, ycoord;
         UBYTE inside;

         Args = (struct acUserMovement *)NotifyArgs->Args;

         /*** Reset the timer and motion sensitivity counter ***/

         Self->MotionTimer = 0;
         UnsubscribeTimer(Self->Head.UniqueID);

         /*** Kill any existing tool-tip due to pointer movement ***/

         if (Self->ToolTipID) {
            ActionMsg(AC_Free, Self->ToolTipID, NULL);
            Self->ToolTipID = NULL;
         }

         /*** Determine whether we are inside or outside the clickable area ***/

         inside = TRUE;
         if (Args->OverObject IS Self->DrawableID) {
            if (Self->Dimensions) {
               if (GetFields(Self, FID_Width|TLONG,  &width,
                                   FID_Height|TLONG, &height,
                                   FID_XCoord|TLONG, &xcoord,
                                   FID_YCoord|TLONG, &ycoord,
                                   TAGEND) IS ERR_Okay) {
                  if ((Args->XCoord < xcoord) OR (Args->YCoord < ycoord) OR
                      (Args->XCoord >= xcoord + width) OR (Args->YCoord >= ycoord + height)) {
                     inside = FALSE;
                  }
               }
            }
         }
         else inside = FALSE;

         if (inside) {
            if (Self->Pointer[0]) {
               if (SetCursor(NULL, ACF_BUFFER, 0, Self->Pointer, Self->Head.UniqueID, NULL) IS ERR_Okay) {
                  Self->PointerLocked = TRUE;
               }
            }

            if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
            else if (Self->State != STATE_INSIDE) {
               Self->State = STATE_ENTERED;
            }

            if ((Self->ToolTip) OR (Self->Flags & OCF_AUTOACTIVATE)) {
               SubscribeTimer(Self->Head.UniqueID, 100);
            }
         }
         else {
            Self->Flags &= ~OCF_ACTIVATED; /* Activation recording is reset when the pointer moves away */

            if (Self->State IS STATE_EXITED) return(ERR_Okay);
            else {
               Self->State = STATE_EXITED;

               if ((Self->PointerLocked) AND (!(Self->Flags & OCF_CLICKED))) {
                  RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
                  Self->PointerLocked = FALSE;
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Calling this action on an onclick object will send Activate messages to its children.
###END###

*****************************************************************************/

static ERROR ONCLICK_Activate(struct OnClick *Self, APTR Args)
{
   LONG count;
   WORD i;

   Self->Flags |= OCF_ACTIVATED;

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
         if (Self->Flags & OCF_MESSAGE) {
            for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
         }
         else for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** OnClick: ClosingTag
*/

static ERROR ONCLICK_ClosingTag(struct OnClick *Self, APTR Args)
{
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Temporarily disables an onclick object.
###END###

*****************************************************************************/

static ERROR ONCLICK_Disable(struct OnClick *Self, APTR Args)
{
   Self->Flags |= OCF_DISABLED;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Enables the functionality of an onclick object.
###END###

*****************************************************************************/

static ERROR ONCLICK_Enable(struct OnClick *Self, APTR Args)
{
   Self->Flags &= ~OCF_DISABLED;
   return(ERR_Okay);
}

/*****************************************************************************
** OnClick: Free
*/

static ERROR ONCLICK_Free(struct OnClick *Self, APTR Args)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   /*** Remove all subscriptions to our drawable container ***/

   if (Self->DrawableID) {
      OBJECTPTR drawable;
      if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
         UnsubscribeAction(drawable, NULL, GetUniqueID(Self));
         ReleaseObject(drawable);
      }
   }

   if (Self->PointerLocked) {
      RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
      Self->PointerLocked = FALSE;
   }

   if (Self->ToolTip) {
      FreeMemory(Self->ToolTip);
      Self->ToolTip = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** OnClick: Init
*/

static ERROR ONCLICK_Init(struct OnClick *Self, APTR Void)
{
   OBJECTID OwnerID;
   OBJECTPTR drawable;

   if (!Self->DrawableID) {
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      SubscribeActionTags(drawable, Self, AC_Free, TAGEND);
      ReleaseObject(drawable);
   }

   SubscribeDrawable(Self, Self->DrawableID);

   return(ERR_Okay);
}

/****************************************************************************/

static void SubscribeDrawable(struct OnClick *Self, OBJECTID ObjectID)
{
   OBJECTPTR drawable;
   LONG count, i;

   if (AccessObject(ObjectID, 4000, &drawable) IS ERR_Okay) {
      SubscribeActionTags(drawable, Self,
         AC_UserClick,
         AC_UserClickRelease,
         AC_UserMovement,
         TAGEND);
      ReleaseObject(drawable);

      /*** Subscribe to child drawables in case they are clicked ***/

      if ((count = TotalChildren(ObjectID))) {
         struct ChildEntry list[count];
         if (ListChildren(ObjectID, list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) {
               if (GetClassID(list[i].ObjectID) IS ID_RENDER) {
                  SubscribeDrawable(Self, list[i].ObjectID);
               }
            }
         }
      }
   }
}

/*****************************************************************************
** OnClick: NewObject
*/

static ERROR ONCLICK_NewObject(struct OnClick *Self, APTR Args)
{
   Self->Button = 1;
   Self->Delay = 0.75;
   return(ERR_Okay);
}

/*****************************************************************************
** OnClick: Timer
*/

static ERROR ONCLICK_Timer(struct OnClick *Self, struct acTimer *Args)
{
   OBJECTPTR script;
   OBJECTID *list;
   WORD i;

   if ((Self->MotionTimer < Self->Delay * 1000) AND (Self->MotionTimer + Args->TimeElapsed >= Self->Delay * 1000)) {
      if (Self->Flags & OCF_AUTOACTIVATE) {
         if ((Self->Flags & OCF_AUTOLIMIT) AND (Self->Flags & OCF_ACTIVATED));
         else acActivate(Self);
      }

      if (Self->ToolTipID) {
         acFreeID(Self->ToolTipID);
         Self->ToolTipID = NULL;
      }

      if (Self->ToolTip) {
         if (NewObject(ID_SCRIPT, NF_CHILD, &script, NULL) IS ERR_Okay) {

            SetFields(script, FID_Location|TSTRING, "templates:tooltip",
                              TAGEND);

            SetUnlistedField(script, "Text", Self->ToolTip);

            if (acInit(script) IS ERR_Okay) {
               acActivate(script);

               if ((GetField(script, FID_ObjectList, FT_POINTER, &list) IS ERR_Okay) AND (list)) {
                  for (i=0; list[i]; i++) {
                     if ((CheckObjectExists(list[i], NULL)) IS ERR_Okay) {
                        Self->ToolTipID = list[i];
                        break;
                     }
                  }
               }

               if (!Self->ToolTipID) {
                  if ((GetField(script, FID_PrimaryObject, FT_LONG, &Self->ToolTipID) != ERR_Okay) OR (!Self->ToolTipID)) {
                     DPrintF("@OnClick:","[Timer] Failed to retrieve the object hook for the tool tip.");
                  }
               }
            }

            acFree(script);
         }
      }
   }

   Self->MotionTimer += Args->TimeElapsed;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Button
Short:  Determines the button that the OnClick object will react to.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can set the Button field to the button number that the OnClick object
should react to.  A value of 1 is used for the left button, 2 for the right
button and 3 for the middle button.  If the pointing device has more than
3 buttons and the software supports it, further numbers are assigned to each
button in a device-associated order.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Determines the drawable to use for click-monitoring and frame setting.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create an OnClick object it should be
contained by a drawable area, such as a Render object.  However there may
be situations where you want to initialise an OnClick object to a
non-graphical container.  If this is the case then you must set the
Drawable field so that the OnClick object knows what object it needs to
monitor for mouse clicks.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags that modify onclick behaviour.
Type:   LONG/FLAGS
Prefix: OCF
Status: Read/Write

###DESCRIPTION###
<p>Special flags supported by the onclick object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">AUTOACTIVATE</td><td>
If this flag is set then the onclick object will automatically activate itself if the pointer enters the area and stops.
Once a preset amount of time has elapsed and the pointer has remained still, the onclick object will activate itself (as
if the user had clicked a mouse button).
</td></tr>
<tr><td width="10%">DISABLED</td><td>
When an onclick object has been disabled, this flag will be automatically
set to indicate that it is not functioning.  You can disable an onclick object
by calling the Disable action, while the Enable action will reactivate a disabled onclick object.
</td></tr>
<tr><td width="10%">DOUBLECLICK</td><td>
When this flag is enabled, an onclick object will only respond to double clicks.  Single clicks are always ignored.
</td></tr>
<tr><td width="10%">HELDRELEASE</td><td>
<p>If this flag is set, the onclick object will activate its children
as soon as the mouse button is released.  The only restrction is that the
mouse <i>must have moved</i> from the position where the original click took
place.</p>
<p>The flag is an equally opposite complement to the RELEASE flag, so
if you set both flags the children will be activated on any condition
of a button's release.</p>
</td></tr>
<tr><td width="10%">IGNOREDOUBLE</td><td>
<p>Set this flag if you want to limit the onclick object to responding to
single-clicks only.  Any secondary clicks that occur within the double-click
timeout period will be ignored by the object.</p>
</td></tr>
<tr><td width="10%">MESSAGE</td><td>
<p>Setting this flag will force the onclick object to always use delayed
messaging when activating child objects, rather than attempting to use
the faster method of instant object activation.</p>
</td></tr>
<tr><td width="10%">RELEASE</td><td>
<p>If this flag is set, the onclick object will activate its children
as soon as the mouse button is released.  The only restrction is that the
mouse must still be in the same position from when the original click
took place.  If any movement has occured then the children will not be
activated.</p>
<p>The flag is an equally opposite complement to the HELDRELEASE flag, so
if you set both flags the children will be activated on any condition
of a button's release.</p>
</td></tr>
<tr><td width="10%">TOGGLEFRAME</td><td>
If this flag is set then the onclick object will toggle between its
drawable's current frame and the value in the
<a href="#ReleaseFrame">ReleaseFrame</a> field whenever a user click is
detected.  Button releases will not have an effect on the drawable's frame.
</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Provides an easy method for frame-switching when a button press is detected.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you would like an onclick object to change the frame number of its
related drawable when a click is detected, set this field to the frame
number that you want to use.  Note that the default behaviour is to change
the drawable's frame back to 0 when the button is released.  You can
change this behaviour by setting the <a href="#ReleaseFrame">ReleaseFrame</a>
field or the TOGGLEFRAME flag.</p>

###SEE ALSO###
Field: Drawable, ReleaseFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of an onclick area.
Type:   DOUBLE/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>If you want an OnClick object to monitor a specific region of the drawable
it is connected to, set the XCoord, YCoord, Width and Height fields.  If you
would rather have the OnClick object monitor the entire area of the drawable,
do not set any of the fields.</p>

<p>To set a relative height, use the FD_PERCENT flag when setting the field;
otherwise a fixed height is assumed.</p>

###SEE ALSO###
Field: Width

###END###

*****************************************************************************/

static ERROR GET_Height(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerHeight, YCoord, YOffset;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FD_FLOAT, &ContainerHeight);
            GetFields(Self, FID_YCoord|TFLOAT,  &YCoord,
                            FID_YOffset|TFLOAT, &YOffset,
                            TAGEND);
            Value->Double = ContainerHeight - YOffset - YCoord;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else Value->Double = 0;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height / ContainerHeight * 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FD_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
            else YCoord = Self->YCoord * 100 / ContainerHeight;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
            else YOffset = Self->YOffset * 100 / ContainerHeight;
            Value->Double = (ContainerHeight - YOffset - YCoord) / ContainerHeight * 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[OnClick] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[OnClick] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Pointer
Short:  Defines the pointer image to use when the pointer enters the vicinity of the onclick area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Set the Pointer field if you would like to use a special pointer image
when the user's pointer enters the onclick area.  Valid image names for the
pointer are defined in the Pointer class documentation.</p>

###SEE ALSO###
Class: Pointer

###END###

*****************************************************************************/

static ERROR GET_Pointer(struct OnClick *Self, STRING *Value)
{
   *Value = Self->Pointer;
   return(ERR_Okay);
}

static ERROR SET_Pointer(struct OnClick *Self, STRING Value)
{
   WORD i;
   for (i=0; (i < LEN_POINTER-1) AND (Value[i]); i++) Self->Pointer[i] = Value[i];
   Self->Pointer[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ReleaseFrame
Short:  If set, an onclick object will change to this frame number when a click is released.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>If this field is set to a frame value, an onclick object will
automatically change the frame of its drawable to the indicated value when
a click is released.  This behaviour only occurs when this field is used
in conjunction with the <a href="#Frame">Frame</a> field.</p>

###SEE ALSO###
Field: Frame

###END###

*****************************************************************************/

static ERROR SET_ReleaseFrame(struct OnClick *Self, LONG *Value)
{
   Self->ReleaseFrame = *Value;
   Self->Flags |= OCF_RELEASEFRAME;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ToolTip
Short:  Set this field with a string to enable a delayed tool tip within the onclick area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>This field provides tool tip support for onclick objects.  Tool tips are
enabled when the mouse pointer sits over an onclick area without
motion.  When this occurs, a small box will pop up containing the text that
you have set in the ToolTip field.  When the pointer is moved, the tool tip
disappears.</p>

<p>The time that it takes for a tool tip to appear is controlled
internally and may be adjusted only by the user.</p>

###END###

*****************************************************************************/

static ERROR SET_ToolTip(struct OnClick *Self, STRING Value)
{
   WORD i;

   if (Self->ToolTip) { ReleaseMemory(Self->ToolTip); Self->ToolTip = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->ToolTip, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ToolTip[i] = Value[i];
         Self->ToolTip[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of an onclick area.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If you want an onclick object to monitor a specific region of the drawable
that it is connected to, set the XCoord, YCoord, Width and Height fields.  If
you would rather have the onclick object monitor the entire area of the
drawable, do not set any of the fields.</p>

<p>To set a relative width, use the FD_PERCENT flag when setting the field;
otherwise a fixed width is assumed.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width * ContainerWidth / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FD_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = Self->XCoord * 100 / ContainerWidth;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
            else XOffset = Self->XOffset * 100 / ContainerWidth;
            Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_Width(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     XCoord
Synonyms: X
Short:    The horizontal position of the OnClick area.
Type:     DOUBLE/PERCENTAGE
Status:   Get/Set

###DESCRIPTION###
<p>If you want an OnClick object to monitor a specific region of the drawable
it is connected to, set the XCoord, YCoord, Width and Height fields.  If you
would rather have the OnClick object monitor the entire area of the drawable,
do not set any of the fields.</p>

<p>To set a relative/percentage based value, you must use the FD_PERCENT flag
or the value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerWidth, Width;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_X) Value->Double = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XCoord * ContainerWidth / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - Width - Self->XOffset;
            else Value->Double = ContainerWidth - Width - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
      Self->XCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
      Self->XCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Specifies the horizontal offset of the OnClick area.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the clickable area will be
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the area is dynamically
calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the clickable area will start from an X coordinate calculated from the
formula &quot;XCoord = ContainerWidth - Width - XOffset&quot;.</p>

###SEE ALSO###
Field: Width, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerWidth, Width;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_XOFFSET) {
         Value->Double = Self->XOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XOffset * ContainerWidth / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_X) Value->Double = ContainerWidth - (Self->XCoord + Width);
            else Value->Double = ContainerWidth - ((Self->XCoord * ContainerWidth / 100) + Width);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:     YCoord
Synonyms: Y
Short:    The vertical position of the OnClick area.
Type:     DOUBLE/PERCENTAGE
Status:   Get/Set

###DESCRIPTION###
<p>If you want an OnClick object to monitor a specific region of the drawable
that it is connected to, set the XCoord, YCoord, Width and Height fields.  If
you would rather have the OnClick object monitor the entire area of the
drawable, do not set any of the fields.</p>

<p>To set a relative/percentage based value, you must use the FD_PERCENT flag
or the value will be interpreted as fixed.  Negative values are permitted.</p

###SEE ALSO###
Fields: Height, XCoord, XOffset, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerHeight, Height;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_Y) {
         Value->Double = Self->YCoord;
      }
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YCoord * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) Value->Double = ContainerHeight - Height - Self->YOffset;
            else Value->Double = ContainerHeight - Height - (ContainerHeight * Self->YOffset / 100);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {

   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
      Self->YCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
      Self->YCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Specifies the vertical offset of the OnClick area.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the clickable area will be
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the area is dynamically
calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the clickable area will start from a Y coordinate calculated from the
formula &quot;YCoord = ContainerHeight - Height - YOffset&quot;.</p>

###SEE ALSO###
Field: Height, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct OnClick *Self, struct Variable *Value)
{
   OBJECTPTR Drawable;
   FLOAT ContainerHeight, Height;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_YOFFSET) {
         Value->Double = Self->YOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YOffset * ContainerHeight / 100;
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (AccessObject(Self->DrawableID, 1000, &Drawable) IS ERR_Okay) {
            GetField(Drawable, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_Y) Value->Double = ContainerHeight - (Self->YCoord + Height);
            else Value->Double = ContainerHeight - ((Self->YCoord * ContainerHeight / 100) + Height);
            ReleaseObject(Drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_YOffset(struct OnClick *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

